/* ---------------------------------------------------------------- */
//
// Andor library (TCP)
// 2016-10-25  Christoph C. Birk (birk@carnegiescience.edu)
// 2018-05-14  TCP/IP version
// 2019-04-17  use (float) for rolling average buffer
//
/* ---------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <unistd.h>

#include "andor.h"
#include "andorTCP.h"
#include "tcpip.h"
#include "ptlib.h"
#include "utils.h"

/* ---------------------------------------------------------------- */

#define DEBUG           1
#define TIME_TEST       1

#ifndef PREFUN
#define PREFUN          __func__
#endif

/* ---------------------------------------------------------------- */
/* andor_() */
/* ---------------------------------------------------------------- */
 
AndorStruct* andor_create(const char* host,int port)
{
  int i;

  AndorStruct *self = (AndorStruct*)malloc(sizeof(AndorStruct));
  if (!self) return NULL;

  strcpy(self->host,host);
  self->port = port;
  self->handle = -1;                   /* connection closed */
  self->err = 0;                       /* used from threads */
  self->serverVersion[0] = '\0';
  self->serialNumber[0] = '\0';
  self->tempSensor = 0.0f;
  strcpy(self->tempStatus,"unknown");
  self->aoiW = self->aoiH = 0;
  self->expTime = 0.5f;
  self->fps = 0.0;
  self->rolling = 0;

  pthread_mutex_init(&self->ioLock,NULL);
  pthread_mutex_init(&self->frameLock,NULL);

  self->seqNumber = 0;
  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *frame = &self->frames[i];
    frame->data = NULL;
  }
  self->tid = 0;
  self->stop_flag = 1;

  return self;
}

/* ---------------------------------------------------------------- */

static int andor_request(AndorStruct* self,const char* cmd,char* res,int tout)
{
  char command[128],answer[128];
  assert(strlen(cmd) < sizeof(command)+2);
  assert(self->handle >= 0);
  if (self->handle < 0) return E_NOTINIT;
  assert(pthread_mutex_trylock(&self->ioLock));  /* must fail here */

  sprintf(command,"%s\n",cmd);
  int err = TCPIP_Request3(self->handle,command,answer,sizeof(answer),tout);
  if (!err) { if (res) strcpy(res,answer); }

  return err;
}

/* ---------------------------------------------------------------- */

int andor_connect(AndorStruct* self)
{
  int  err=0,handle=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s:%d)\n",PREFUN,self->host,self->port);
#endif
  assert(self);
  assert(self->handle < 0);

  pthread_mutex_lock(&self->ioLock);
  if (!err) {
    handle = TCPIP_CreateClientSocket(self->host,self->port,&err);
#if (DEBUG > 0)
    printf("%s(%s:%d): handle=%d\n",PREFUN,self->host,self->port,handle);
#endif
    if (!err) self->handle = handle;
  }
  if (!err) { char buf[128];
    err = andor_request(self,"version",buf,5);
    if (!err) {
      assert(strlen(buf) < sizeof(self->serverVersion));
      sscanf(buf,"%s %u",self->serverVersion,&self->cookie);
      sprintf(buf,"serverVersion= %s",self->serverVersion);
      message(self,buf,MSG_FILE);
    }
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_open(AndorStruct* self)
{
  int  err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s:%d)\n",PREFUN,self->host,self->port);
#endif
  assert(self);

  if (self->handle < 0) err = andor_connect(self);

  if (!err) { char buf[128];
    pthread_mutex_lock(&self->ioLock); /* slow AT_Open() at Minnow */
    err = andor_request(self,"serial",buf,40);
    if (!err) {
      assert(strlen(buf) < sizeof(self->serialNumber));
      strcpy(self->serialNumber,buf);
      sprintf(buf,"serialNumber= %s",self->serialNumber);
      message(self,buf,MSG_FILE);
    } 
    pthread_mutex_unlock(&self->ioLock);
  }

  return err;
}

/* ---------------------------------------------------------------- */

int andor_setup(AndorStruct* self,int dim,int bin)
{ 
  int  err=0;
  char cmd[128],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,self->serialNumber,dim,bin);
#endif
  assert(self);
  assert(self->handle >= 0);
  if (self->handle < 0) return E_NOTINIT;
  assert(self->tid == 0);
  if (self->tid) return E_RUNNING;

  pthread_mutex_lock(&self->ioLock);
  sprintf(cmd,"offtime %ld",cor_time(0));
  if (!err) err = andor_request(self,cmd,buf,5);
#if (DEBUG > 1)
  // NOTE: timeout might be too short if not "open" yet
  if (!err) err = andor_request(self,"state",buf,5);
  printf("%s: state= %s\n",PREFUN,buf); 
#endif
  sprintf(cmd,"setup %d %d",dim,bin);
  if (!err) err = andor_request(self,cmd,buf,5);
  if (!err) self->aoiW = self->aoiH = dim/bin;
  sprintf(cmd,"exptime");
  if (!err) err = andor_request(self,cmd,buf,5);
  if (!err) self->expTime = atof(buf); 
  pthread_mutex_unlock(&self->ioLock);
 
  self->err = err;

  return err;  
}

/* ---------------------------------------------------------------- */

int andor_disconnect(AndorStruct* self)
{
  int err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self);

  if (self->tid) return E_RUNNING;

  pthread_mutex_lock(&self->ioLock);
  if (self->handle >= 0) err = close(self->handle);
  self->handle = -1;
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_close(AndorStruct* self)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s %s(%p): tid=%ld\n",__FILE__,PREFUN,self,self->tid);
#endif
  assert(self);

  if (self->handle < 0) return 0;
  if (self->tid) return E_RUNNING;

  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,"close",NULL,5);
  pthread_mutex_unlock(&self->ioLock);
  andor_disconnect(self);
  message(self,"disconnected from andorserver",MSG_FLUSH);

  return err;
}

/* ---------------------------------------------------------------- */

void andor_free(AndorStruct* self)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self);

  pthread_mutex_destroy(&self->ioLock);
  pthread_mutex_destroy(&self->frameLock);

  free((void*)self);
}

/* ---------------------------------------------------------------- */

int andor_temperature(AndorStruct *self)
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif
  assert(self->handle >= 0);

  pthread_mutex_lock(&self->ioLock); 
  int err = andor_request(self,"temperature",buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) { double value=0; char buf1[64]="",buf2[64]="";
    int n = sscanf(buf,"%lf %s %s",&value,buf1,buf2);
    if (n >= 1) self->tempSensor = value;
    if (n == 2) strcpy(self->tempStatus,buf1);
    if (n == 3) sprintf(self->tempStatus,"%s %s",buf1,buf2);
  }

  return err;
}

/* ---------------------------------------------------------------- */

int andor_tempcon(AndorStruct* self,int value)
{
  char cmd[128],buf[128];

  sprintf(cmd,"tempcon %d",value);
  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_exptime(AndorStruct* self,double exptime)
{
  char cmd[128],buf[128];

  exptime = fmin(MAX_EXPTIME,fmax(0.01,exptime));
  sprintf(cmd,"exptime %.3f",exptime);
  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) self->expTime = exptime;

  return err;
}

/* ---------------------------------------------------------------- */

#if 0 // at TCP or USB ? -- here -- don't send it to server
int andor_rolling(AndorStruct* self,int rolling)
{
  char cmd[128],buf[128];

  sprintf(cmd,"rolling %d",rolling);
  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,cmd,buf,5);
  pthread_mutex_unlock(&self->ioLock);
  if (!err) self->rolling = rolling;

  return err;
}
#endif

/* ---------------------------------------------------------------- */

static void* run_cycle(void* param)
{
  AndorStruct *self=(AndorStruct*)param;
  int     i,n,dim,err=0,last_err=0;
  u_int   seq;
  double  t1,t2,tst,tmp;
  char    cmd[128],buf[128];
  float   *roll_buf=NULL;              /* NEW v0300 */
#if (TIME_TEST > 0)
  double s1=0,s2=0,sn=0,sm=0,dt;
  time_t last=0;
#endif
#if (DEBUG > 0)
  fprintf(stderr,"%s: %s(%p)\n",__FILE__,PREFUN,param);
#endif
  assert(self->handle >= 0);

  message(self,PREFUN,MSG_FLUSH);

  t1 = walltime(0);

  int npix = self->aoiW * self->aoiH;
  int nbytes = npix * sizeof(u_short);
  u_char *data = (u_char*)malloc(nbytes);

  self->fps = 0;
  self->err = 0;
  while (!self->stop_flag) {
    msleep(5);
    sprintf(cmd,"next %.1f",fmin(self->expTime+1.0,2.0));
    pthread_mutex_lock(&self->ioLock);
    err = andor_request(self,cmd,buf,5);
    if (err) {
      fprintf(stderr,"%s: err=%d\n",PREFUN,err);
    } else
    if (!strncmp(buf,"-E",2)) {
      if (strstr(buf,"-Enodata")) {
        float dt = walltime(0)-t1;
        if (dt > self->expTime+0.5f) {       /* +0.5 v0221 */
          fprintf(stderr,"%s: %s:%d - no data %.1f\n",
                  get_local_timestr(buf,0),self->host,self->port,dt);
        }
      } else {
        printf("%s: other error '%s'\n",PREFUN,buf);
        char *p = strchr(buf,'='); 
        err = (p) ? atoi(p+1) : E_ERROR;
      }
    } else {
      n = sscanf(buf,"%u %d %lf %lf",&seq,&dim,&tst,&tmp);
      if (n != 4) message(self,buf,MSG_FLUSH);
      assert(n >= 4);
      self->tempSensor = (float)tmp;
      assert(dim == self->aoiW);
      assert(dim == self->aoiH);
      for (i=0,n=0; i<nbytes/42; i++) {
        ssize_t r = TCPIP_Recv(self->handle,(char*)data+n,nbytes-n,2);
        if (r <= 0) break;
        n += r; if (n >= nbytes) break;
      }
      t2 = walltime(0);
      self->fps = 0.7*self->fps + 0.3/(t2-t1);
      t1 = t2;
      if (n != nbytes) {               /* data missing */
        err = E_INCFRAME;
      } else {
#if (DEBUG > 1)
        if (seq > self->seqNumber+1) {
          sprintf(buf,"skipped %u frames",seq-self->seqNumber-1);
          message(self,buf,MSG_WINDOW);
        }
#endif
        self->seqNumber = seq;
        AndorFrame *frame = andor_frame4writing(self,seq);
        if (frame) {
          frame->seqNumber = seq;
          frame->timeStamp = tst;
          if (self->rolling == 0) {    /* rolling average */
            memcpy(frame->data,data,nbytes);
            if (roll_buf) { free((void*)roll_buf); roll_buf=NULL; }
          } else { register u_short *d; register float *p;
#if (TIME_TEST > 0) // IDEA "debug" level
            double c1 = walltime(0);
#endif
            if (!roll_buf) {           /* first frame */
              roll_buf = (float*)malloc(npix*sizeof(float));
              d = (u_short*)data; p = roll_buf;
              for (i=0; i<npix; i++,p++,d++) { *p = (float)*d; }
#if (TIME_TEST > 0)
              s1 = s2 = sn = sm = 0;
#endif
            } else { float mul=self->rolling,div=1+self->rolling;
              d = (u_short*)data; p = roll_buf;
              for (i=0; i<npix; i++,p++,d++) {
                *p = ((*p * mul) + (float)*d) / div;
              }
            }
            d = frame->data; p = roll_buf;
            for (i=0; i<npix; i++,p++,d++) { *d = (u_short)(*p+0.5f); }
#if (TIME_TEST > 0)
            dt = (walltime(0)-c1)*1000.0;   /* [ms] */
            if (sm != 0) { s1 += dt; s2 += dt*dt; sn += 1.0; }
            if (dt > sm) sm = dt;
            if ((sn > 0) && (cor_time(0) > last)) { 
              double ave = s1/sn;
              double sig = sqrt(s2/sn-ave*ave);
              float *p = (float*)malloc(npix*sizeof(float));
              for (i=0,n=0; i<npix; i+=17) { p[n] = roll_buf[i]; n++; }
              f_sort(p,0,n-1);
              sprintf(buf,"%s(%d): ave=%.1f +- %.2f (max=%.1f)",
                      PREFUN,self->port-SERVER_PORT,ave,sig,sm);
              message(self,buf,MSG_FILE); 
              printf("%s med=%.2f\n",buf,p[n/2]);
              free(p);
              last = cor_time(0);
            }
#endif
          } /* endif(rolling) */
          andor_frame_release(self,frame);
        } // endif(PANIC)
      } // endif(incomplete frame)
    } // endif(err)
    pthread_mutex_unlock(&self->ioLock);
    if (err) {
      if (err != last_err) {
        switch (err) {
        case E_tcpip_send:
          sprintf(buf,"sending req. to '%s' failed",self->host);
          break;
        case E_tcpip_timeout:
          sprintf(buf,"no response from '%s'",self->host);
          break;
        case E_RUNNING:  
          sprintf(buf,"%s: data-acq not running",PREFUN);
          break;
        case E_INCFRAME: 
          sprintf(buf,"incomplete frame from '%s'",self->host); 
          break;
        default:         
          sprintf(buf,"%s: err=%d",PREFUN,err); 
          break;
        }
        message(self,buf,MSG_ERROR);
        last_err = err; 
      }
      msleep(500);
    }
  } /* endwhile (!stop_flag) */

  if (roll_buf) { free((void*)roll_buf); roll_buf=NULL; }
  free((void*)data);
  self->err = err;
#if (DEBUG > 0)
  fprintf(stderr,"%s %s() done\n",__FILE__,PREFUN);
#endif

  return (void*)(long)err;
}

/* ---------------------------------------------------------------- */

int andor_cycle_start(AndorStruct* self)
{
  char buf[128];

  size_t size = self->aoiW * self->aoiH * sizeof(u_short);

  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,"start",buf,5);
  if (!err) andor_request(self,"state",buf,5);
  if (!err) self->seqNumber = stringVal(buf,4);
  pthread_mutex_unlock(&self->ioLock);

  if (!err) { int i;
    for (i=0; i<ANDOR_NBUFS; i++) {
      AndorFrame *frame = &self->frames[i];    
      frame->timeStamp = 0;
      frame->seqNumber = 0;
      assert(!frame->data);
      frame->data = (u_short*)malloc(size);
      assert(((u_long)frame->data & 0x07) == 0);
      frame->w = self->aoiW;
      frame->h = self->aoiH;
      frame->wlock = 0;
      frame->rlock = 0;
    }
  }

  if (!err) {
    self->stop_flag = 0;
    thread_create(run_cycle,(void*)self,&self->tid);
    msleep(50);
  } else {
    self->tid = 0;
  }

  return err;
}

/* ---------------------------------------------------------------- */

int andor_cycle_stop(AndorStruct *self)
{
  int  err=0,i;
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,self);
#endif

  self->stop_flag = 1;
  pthread_join(self->tid,NULL);  
  self->tid = 0;

  pthread_mutex_lock(&self->ioLock);
  andor_request(self,"stop",buf,5);
  andor_request(self,"state",buf,5);
  self->seqNumber = stringVal(buf,4);
  pthread_mutex_unlock(&self->ioLock);

  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *f = &self->frames[i]; 
    while (f->wlock != 0) { 
      msleep(50); fprintf(stderr,"%s: f=%d, wlock=%d\n",PREFUN,i,f->wlock); 
    }
    while (f->rlock != 0) { 
      msleep(50); fprintf(stderr,"%s: f=%d, rlock=%d\n",PREFUN,i,f->rlock); 
    } 
    if (f->data) { free((void*)f->data); f->data = NULL; }
  }

  return err;
}

/* ---------------------------------------------------------------- */

AndorFrame* andor_frame4writing(AndorStruct* self,u_int s)
{
  int i;
  AndorFrame *frame=NULL;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%u)\n",PREFUN,self,s);
#endif

  pthread_mutex_lock(&self->frameLock);
  
  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *f = &self->frames[i];    
    assert(f->data);
    assert(f->wlock == 0);             // no frame locked for writing
    if (f->rlock == 0) {               // not locked for reading
      if (f->seqNumber < s) {          // oldest frame
        frame = f;
        s = frame->seqNumber;
      }
    }
  }
  if (!frame) { char buf[128];
    sprintf(buf,"%s: PANIC: no frame available for writing",__FILE__);
    message(self,buf,MSG_WARN | MSG_FILE);
  } else {
    frame->wlock = 1;
  }
  pthread_mutex_unlock(&self->frameLock);
        
  return frame;
}

/* ---------------------------------------------------------------- */

AndorFrame* andor_frame4reading(AndorStruct* self,u_int s)
{
  int i;
  AndorFrame *frame=NULL;

  if (self->stop_flag) return NULL;

  pthread_mutex_lock(&self->frameLock);
  
  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *f = &self->frames[i];    
    assert(f->data);
    if (f->wlock == 0) {               // not locked for writing
      if (f->seqNumber > s) {          // newest frame
        frame = &self->frames[i];
        s = frame->seqNumber;
      }
    }
  }
  if (frame) {
    frame->rlock += 1;
  }
  pthread_mutex_unlock(&self->frameLock);
        
  return frame;
}


/* ---------------------------------------------------------------- */

void andor_frame_release(AndorStruct* self,AndorFrame *frame)
{
  pthread_mutex_lock(&self->frameLock);

  if (frame->wlock) {                  /* is locked for writing */
    assert(frame->rlock == 0);
    assert(frame->wlock == 1);
    frame->wlock = 0;
  } else {                             /* is locked for reading */
    assert(frame->rlock >  0);
    frame->rlock -= 1;
    assert(frame->rlock >= 0);
  }
  
  pthread_mutex_unlock(&self->frameLock);
}

/* ---------------------------------------------------------------- */

int andor_server(AndorStruct* self,const char* cmd,char* res)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s,%s)\n",PREFUN,self,cmd,par);
#endif
  assert(self);
  if (self->handle < 0) return E_NOTINIT;

  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,cmd,res,5);
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_sim(AndorStruct* self,const char* command,char* res)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s,%s)\n",PREFUN,self,cmd,par);
#endif
  assert(self);
  if (self->handle < 0) return E_NOTINIT;

  pthread_mutex_lock(&self->ioLock);
  int err = andor_request(self,command,res,5);
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

