/* ---------------------------------------------------------------- */
//
// Andor library (USB)
// 2016-10-25  Christoph C. Birk (birk@carnegiescience.edu)
//
// Note: recovery after power-failure is not possible because
//       the driver just crashes when AT_Close() or anything
//       else is called.
//       Rebooting the PC is the only safe way to recover from
//       a cable disconnect or power failure.
//
/* ---------------------------------------------------------------- */

#define DEBUG           1
#define TIME_TEST       0

/* ---------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <math.h>
#include <assert.h>

#include "andor.h"
#include "andorUSB.h"
#include "ptlib.h"
#include "utils.h"
#include "random.h"

/* ---------------------------------------------------------------- */

#ifndef PREFUN
#define PREFUN          __func__
#endif

#define WSTR_SERIAL     L"SerialNumber"
#define WSTR_MODEL      L"CameraModel"
#define WSTR_NAME       L"CameraName"
#define WSTR_SENSOR_W   L"SensorWidth"
#define WSTR_SENSOR_H   L"SensorHeight"
#define WSTR_PIXEL_H    L"PixelHeight"
#define WSTR_PIXEL_W    L"PixelWidth"
#define WSTR_RTIME      L"ReadoutTime"
#define WSTR_COOLING    L"SensorCooling"

#define WSTR_BINH       L"AOI HBin"
#define WSTR_BINV       L"AOI VBin"
#define WSTR_BINNING    L"AOI Binning"
#define WSTR_AOI_W      L"AOI Width"
#define WSTR_AOI_H      L"AOI Height"
#define WSTR_AOI_L      L"AOI Left"
#define WSTR_AOI_T      L"AOI Top"

#define WSTR_STEMP      L"SensorTemperature"
#define WSTR_TSTAT      L"TemperatureStatus"

#define WSTR_EXPTIME    L"ExposureTime"
#define WSTR_START      L"Acquisition Start"

/* ---------------------------------------------------------------- */

#if (DEBUG > 1) && defined(SIM_ONLY)
static const char* wcs2char(const wchar_t *wbuf)
{
  static char buf[512];
  wcstombs(buf,wbuf,sizeof(buf));
  return buf;
}
#endif

/* ---------------------------------------------------------------- */

static pthread_mutex_t andor_mutex;
static volatile int andor_initialized=0;

#ifdef SIM_ONLY
#define MAX_NANDORS 3
#define SENSOR_G 2048
static int sim_mode[MAX_NANDORS]={0,1,0};  /* 0:imaging 1:SH */
static int sim_aoi_w[MAX_NANDORS]={SENSOR_G,SENSOR_G,SENSOR_G};
static int sim_aoi_h[MAX_NANDORS]={SENSOR_G,SENSOR_G,SENSOR_G};
static int sim_aoi_l[MAX_NANDORS]={1,1,1};
static int sim_aoi_t[MAX_NANDORS]={1,1,1};
static double sim_temp[MAX_NANDORS]={25.0,25.0,25.0};
static AT_BOOL sim_tcon[MAX_NANDORS]={0,0,0};
static double sim_exptime[MAX_NANDORS]={0.5,0.5,0.5};
static AT_64 sim_hbin[MAX_NANDORS]={1,1,1};
static AT_64 sim_vbin[MAX_NANDORS]={1,1,1};
static u_char* sim_buffer[MAX_NANDORS]={NULL,NULL,NULL};
static int sim_size[MAX_NANDORS]={0,0,0};
static double sim_walltime[MAX_NANDORS];
static double sim_flux[MAX_NANDORS]={1.0e5,1.0e5,1.0e5};
static double sim_sigma[MAX_NANDORS]={3.0,3.0,3.0};
static double sim_radius[MAX_NANDORS]={2.0,2.0,2.0};

int AT_InitialiseLibrary(void) 
{ 
#if (DEBUG > 0)
  fprintf(stderr,"%s\n",__func__); 
#endif
  msleep(1000); 
  return 0; 
}

int AT_FinaliseLibrary(void) 
{ 
#if (DEBUG > 0)
  fprintf(stderr,"%s\n",__func__); 
#endif
  return 0; 
}

int AT_Open(int i,AT_H *h) 
{
  if (i < 0) return -1;
  if (i >= MAX_NANDORS) return -1;
  *h = i; 
  sim_aoi_w[i] = sim_aoi_h[i] = SENSOR_G;
  sim_aoi_l[i] = sim_aoi_t[i] = 1;
  // slow at MinnowBoard sleep(5);
  return 0;
}
int AT_Close(AT_H h) { return 0; }

int AT_GetBool(AT_H h,const AT_WC* f,AT_BOOL* v) { *v=1; return 0; }
int AT_SetBool(AT_H h,const AT_WC* f,AT_BOOL v) 
{ 
  if (h < 0) return 1;
  if (!wcscmp(f,WSTR_COOLING)) sim_tcon[h] = v;
  return 0; 
}

int AT_GetString(AT_H h,const AT_WC* f,AT_WC* s,int z)
{ 
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_SERIAL)) swprintf(s,32,L"sim%04d",1+h);
  else if (!wcscmp(f,WSTR_MODEL))  wcscpy(s,L"simModel");
  else if (!wcscmp(f,WSTR_NAME))   wcscpy(s,L"simName");
  else                             wcscpy(s,L"unknown");
  return 0; 
}

int AT_GetInt(AT_H h,const AT_WC* f,AT_64* v) 
{ 
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_SENSOR_W))  *v = 2048;
  else if (!wcscmp(f,WSTR_SENSOR_H))  *v = 2048;
  else if (!wcscmp(f,WSTR_AOI_W))     *v = sim_aoi_w[h];
  else if (!wcscmp(f,WSTR_AOI_H))     *v = sim_aoi_h[h];
  else if (!wcscmp(f,WSTR_AOI_L))     *v = sim_aoi_l[h];
  else if (!wcscmp(f,WSTR_AOI_T))     *v = sim_aoi_t[h];
  else if (!wcscmp(f,WSTR_BINH))      *v = sim_hbin[h];
  else if (!wcscmp(f,WSTR_BINV))      *v = sim_vbin[h];
  else if (!wcscmp(f,L"AccumulateCount")) *v = 1;
  else if (!wcscmp(f,L"FrameCount"))      *v = 1;
  else                                    *v = 0;
  return 0; 
}
int AT_SetInt(AT_H h,const AT_WC* f,AT_64 v) 
{
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_BINH))  sim_hbin[h] = v;
  else if (!wcscmp(f,WSTR_BINV))  sim_vbin[h] = v;
  else if (!wcscmp(f,WSTR_AOI_W)) sim_aoi_w[h] = v;
  else if (!wcscmp(f,WSTR_AOI_H)) sim_aoi_h[h] = v;
  else if (!wcscmp(f,WSTR_AOI_L)) sim_aoi_l[h] = v;
  else if (!wcscmp(f,WSTR_AOI_T)) sim_aoi_t[h] = v;

  return 0; 
}

int AT_SetFloat(AT_H h,const AT_WC* f,double v) 
{ 
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_EXPTIME)) sim_exptime[h]=v;
  return 0; 
}
int AT_GetFloat(AT_H h,const AT_WC* f,double* v) 
{
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_EXPTIME)) *v=sim_exptime[h];
  else if (!wcscmp(f,WSTR_STEMP)) { 
    if (sim_tcon[h]) sim_temp[h] = (2.0*sim_temp[h] + 0.25)/3.0;
    else             sim_temp[h] = (2.0*sim_temp[h] + 25.0)/3.0;
    *v = GRandom(sim_temp[h],0.05);
  }
  else if (!wcscmp(f,WSTR_PIXEL_W)) *v=6.5;
  else if (!wcscmp(f,WSTR_PIXEL_H)) *v=6.5;
  else if (!wcscmp(f,WSTR_RTIME))   *v=0.009830;
  else                              *v=0.0;
  return 0; 
}
int AT_GetFloatMax(AT_H h,const AT_WC* f,double* v) 
{ 
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_EXPTIME)) *v=30.0;
  return 0; 
}
int AT_GetFloatMin(AT_H h,const AT_WC* f,double* v) 
{ 
  if (h < 0) return 1;
  if      (!wcscmp(f,WSTR_EXPTIME)) *v=0.001;
  return 0; 
}

int AT_IsImplemented(AT_H h,const AT_WC* f,AT_BOOL* m) { *m=1; return 0; }
int AT_IsWritable(AT_H h,const AT_WC* f,AT_BOOL* m) { *m=1; return 0; }

int AT_GetEnumIndex(AT_H h,const AT_WC* f,int* i) 
{ 
  if      (!wcscmp(f,WSTR_TSTAT))   *i = 1;
  else if (!wcscmp(f,WSTR_BINNING)) *i = 2;
  else                              *i = 0; 
#if (DEBUG > 1)
  printf("%s(%s)=%d\n",PREFUN,wcs2char(f),*i);
#endif
  return 0; 
}
int AT_GetEnumStringByIndex(AT_H h,const AT_WC* f,int i,AT_WC* s,int z)
{
#if (DEBUG > 1)
  printf("%s(%s,%d)\n",PREFUN,wcs2char(f),i);
#endif
  if (h < 0) return 1;
  switch(i) {
    case 1:                            /* CCD temperature */
      if (sim_tcon[h]) {
        wcscpy(s,(sim_temp[h] < 0.3) ? L"Stabilised" : L"Cooling"); 
      } else {
        wcscpy(s,L"Cooler Off");
      }
      break;
    case 2:                            /* binning */
      swprintf(s,z,L"%dx%d",sim_hbin[h],sim_vbin[h]); 
      break;
    default: 
      wcscpy(s,L"n/a"); 
      break;
  }
  return 0;
}
int AT_SetEnumString(AT_H h,const AT_WC* f,const AT_WC* s) 
{ 
  char buf[128];
  if (h < 0) return 1;
  wcstombs(buf,s,sizeof(buf));
  if (!wcscmp(f,WSTR_BINNING)) { int a,v;
    if (sscanf(buf,"%dx%d",&a,&v) == 2) { sim_hbin[h] = a; sim_vbin[h] = v; }
  }
  return 0; 
}

int AT_Command(AT_H h,const AT_WC* f) 
{ 
  if (h < 0) return 1;
  if (!wcscmp(f,WSTR_START)) sim_walltime[h] = walltime(0)+0.01+sim_exptime[h];

  return 0; 
}

int AT_QueueBuffer(AT_H h,AT_U8* p,int s) 
{ 
  if (h < 0) return 1;
  sim_buffer[h] = p;
  sim_size[h] = s;
  sim_walltime[h] += 0.01+sim_exptime[h];
  return 0;
}

int AT_WaitBuffer(AT_H h,AT_U8** p,int* s,unsigned int t) 
{
  int i,j,w,x,y,ii,jj; u_short *ubuf; double f,dx,dy;
  if (h < 0) return 1;
  int wait = (int)(1000.0*(sim_walltime[h]-walltime(0)));
  if (wait > (int)t) { msleep(t); return AT_ERR_TIMEDOUT; }
  int npix = sim_size[h]/sizeof(u_short);
  w = (int)sqrt(npix); 
  assert(w*w == npix);
  assert(w == sim_aoi_w[h]);
  double bias = 3.0*sim_hbin[h]*sim_vbin[h];
  double dark = 0.05*sim_exptime[h]*sim_hbin[h]*sim_vbin[h];
  double noise = 3.0*sqrt(sim_hbin[h]*sim_vbin[h])+sqrt(dark);
  u_short *dbuf = (u_short*)malloc(w*sizeof(u_short));
  for (i=0; i<w; i++) { 
    do { f = GRandom(bias,noise); } while (f < -bias);
    dbuf[i] = (u_short)(100+f+0.5);
  }
  ubuf = (u_short*)sim_buffer[h];
  for (i=0,j=0; i<npix; i++,ubuf++) { 
    *ubuf = dbuf[j]; j++; if (j == w) j = IRandom(w);
  }
  free((void*)dbuf);
#if 1 
  double posx[] = { 0.2,0.3,0.4,0.5,0.6,0.7,0.8 };
  double posy[] = { 0.2,0.3,0.4,0.5,0.6,0.7,0.8 };
#else
  double posx[] = { 0.2,0.5,0.8 };
  double posy[] = { 0.2,0.5,0.8 };
#endif
  int nx = (int)(sizeof(posx)/sizeof(double));
  int ny = (int)(sizeof(posy)/sizeof(double));
  double flux = sim_flux[h]*sim_exptime[h];
  if (sim_mode[h]) flux /= nx*ny;
  int dd = (int)(4.0*sim_sigma[h]);
  double f2 = 2.0 * sim_sigma[h] * sim_sigma[h];
  double peak = flux/(f2*3.14);
  static double phi[MAX_NANDORS] = {0,0,0}; 
  phi[h] += (2.0*M_PI)/20.0;
  for (ii=0; ii<nx; ii++) {
    if (!sim_mode[h] && (ii != nx/2)) continue;
    for (jj=0; jj<ny; jj++) {
      if (!sim_mode[h] && (jj != ny/2)) continue;
      double xc = (double)w*posx[ii] + sim_radius[h]*cos(phi[h]);
      double yc = (double)w*posy[jj] + sim_radius[h]*sin(phi[h]);
      ubuf = (u_short*)sim_buffer[h];
      int x1 = imax(0,(int)xc-dd);
      int x2 = imin(w-1,(int)xc+dd);
      int y1 = imax(0,(int)yc-dd);
      int y2 = imin(w-1,(int)yc+dd);
      for (x=x1; x<=x2; x++) {
        dx = xc - x;
        for (y=y1; y<=y2; y++) {
          dy = yc - y;
          f = peak * exp(-(dx*dx+dy*dy)/f2);
          ubuf[x+y*w] += (u_short)(f+0.5);
        }
      }
    }
  }
  wait = (int)(1000.0*(sim_walltime[h]-walltime(0)));
  if (wait > 0) msleep(wait);
  *p = sim_buffer[h];
  *s = sim_size[h];

  return 0;
}

int AT_Flush(AT_H h) { return 0; }

#endif  /* SIM_ONLY */

/* ---------------------------------------------------------------- */

static int get_enum_feature(AT_H h,const AT_WC* f,char* buf,size_t size)
{
  int   i;
  AT_WC wbuf[128];

  int err = AT_GetEnumIndex(h,f,&i);
  if (!err) err = AT_GetEnumStringByIndex(h,f,i,wbuf,128);
  if (!err) wcstombs(buf,wbuf,size);

  return err;
}

static int get_string_feature(AT_H h,AT_WC* f,char* buf,size_t size)
{
  AT_WC wbuf[128];

  int err = AT_GetString(h,f,wbuf,sizeof(wbuf)/sizeof(AT_WC));
  if (!err) wcstombs(buf,wbuf,size);

  return err;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

static pthread_once_t once_control=PTHREAD_ONCE_INIT;

static void andor_initalize(void)      /* protected */
{
#if (DEBUG > 2)
  fprintf(stderr,"%s: enter(%d)\n",PREFUN,once_control);
#endif

  pthread_mutex_init(&andor_mutex,NULL);

#if (DEBUG > 2)
  fprintf(stderr,"%s: leave(%d)\n",PREFUN,once_control);
#endif
}

/* ---------------------------------------------------------------- */
/* andor_() */
/* ---------------------------------------------------------------- */
 
AndorStruct* andor_create(int id)
{
  int i;

  AndorStruct *self = (AndorStruct*)malloc(sizeof(AndorStruct));
  if (!self) return NULL;

  pthread_once(&once_control,andor_initalize);

  self->id = id;
  self->handle = -1;                   /* connection closed */
  self->initialized = 0;
  self->err = 0;
  self->model[0] = '\0';
  self->serialNumber[0] = '\0';
  self->name[0] = '\0';
  self->firmware[0] = '\0';
  self->tempSensor = 0.0f;
  strcpy(self->tempStatus,"unknown");
  self->bias = 0;
  self->sensorW = self->sensorH = 0;
  self->aoiW = self->aoiH = 0;
  self->binH = self->binV = 1;
  self->expTime = 0.5;
  self->readoutTime = 0.0;
  self->fps = 0.0;

  pthread_mutex_init(&self->ioLock,NULL);
  pthread_mutex_init(&self->frameLock,NULL);

  self->seqNumber = 0;
  self->data_1 = self->data_2 = NULL;

  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *frame = &self->frames[i];
    frame->data = NULL;
  }

  self->tid = 0;
  self->stop_flag = 1;

  return self;
}

/* ---------------------------------------------------------------- */

int andor_open(AndorStruct* self)
{
  int     err=0;
  AT_H    handle=0;
  AT_BOOL b=0;
  AT_64   l=0;
  char    buf[128],line[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%s(%d)\n",__FILE__,PREFUN,self->id);
#endif
  assert(self);
  assert(self->handle < 0);

  pthread_mutex_lock(&andor_mutex);
  if (andor_initialized == 0) {
    err = AT_InitialiseLibrary();      /* fast @ minnow */
  }
  if (!err) andor_initialized += 1;
  pthread_mutex_unlock(&andor_mutex);

  if (!err) {
    err = AT_Open(self->id,&handle);   /* slow @ minnow */
#if (DEBUG > 0)
    fprintf(stderr,"%s(%d): handle=%d\n",PREFUN,self->id,handle);
#endif
    if (err) err = E_andor_not_connected;
  }

  if (!err) {
    err = AT_GetBool(handle,L"CameraPresent",&b);
    // returns AT_ERR_NOTIMPLEMENTED(2) if not powered
    if (err || !b) err = E_andor_not_present;
  }
 
  self->model[0] = '\0';
  if (!err) { 
    err = get_string_feature(handle,WSTR_MODEL,buf,sizeof(buf));
    if (!err) {
      sprintf(line,"model= %s",buf);
      message(self,line,MSG_FILE);
      assert(strlen(buf) < sizeof(self->model));
      strcpy(self->model,buf);
    }
  }

  self->serialNumber[0] = '\0';
  if (!err) { 
    err = get_string_feature(handle,WSTR_SERIAL,buf,sizeof(buf));
    if (!err) {
      sprintf(line,"serial= %s",buf);
      message(self,line,MSG_FILE);
      assert(strlen(buf) < sizeof(self->serialNumber));
      strcpy(self->serialNumber,buf);
    } 
  }

  self->name[0] = '\0';
  if (!err) {
    err = get_string_feature(handle,WSTR_NAME,buf,sizeof(buf));
    if (!err) {
#if 0 // redundant with model
      sprintf(line,"name= %s",buf);
      message(self,line,MSG_FILE);
#endif
      assert(strlen(buf) < sizeof(self->name));
      strcpy(self->name,buf);
    }
  }

  self->firmware[0] = '\0';
  if (!err) {
    err = get_string_feature(handle,L"FirmwareVersion",buf,sizeof(buf));
    if (!err) {
      sprintf(line,"firmware= %s",buf); 
      message(self,line,MSG_FILE);
      assert(strlen(buf) < sizeof(self->firmware));
      strcpy(self->firmware,buf);
    }
  }

  if (!err) {
    if (!err) err = AT_GetInt(handle,WSTR_SENSOR_W,&l);
    if (!err) self->sensorW = (int)l;
    if (!err) err = AT_GetInt(handle,WSTR_SENSOR_H,&l);
    if (!err) self->sensorH = (int)l;
#if (DEBUG > 0)
    fprintf(stderr,"sensor= %d,%d\n",self->sensorW,self->sensorH);
#endif
  }

  if (!err) self->handle = handle;

  return err;
}

/* ---------------------------------------------------------------- */

int andor_init(AndorStruct* self)
{ 
  int     err=0;
  AT_H    handle=self->handle;
  AT_BOOL b=0;
  AT_64   l=0;
  char    buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%s(%d)\n",__FILE__,PREFUN,self->id);
#endif
  assert(self);
  assert(self->handle >= 0);

  if (self->initialized) return 0;     /* just once */

  self->bias = 0;
  if (!err) {
    err = AT_GetInt(handle,L"Baseline",&l);
#if (DEBUG > 1)
    if (!err) printf("baseline=%lld\n",l);
#endif
    self->bias = (int)l;
  }

  if (!err) { const AT_WC *feature=WSTR_AOI_W;
    AT_IsImplemented(handle,feature,&b); 
    assert(b);
    AT_IsWritable(handle,feature,&b);
    assert(b);
    err = AT_GetInt(handle,feature,&l);
    if (!err) self->aoiW = l; // bug-fix NEW 2019-01-07
    assert(self->sensorW == l);
  }
  if (!err) { const AT_WC *feature=WSTR_AOI_H;
    AT_IsImplemented(handle,feature,&b);
    assert(b);
    AT_IsWritable(handle,feature,&b);
    assert(b);
    err = AT_GetInt(handle,feature,&l);
    if (!err) self->aoiH = l;
    assert(self->sensorH == self->aoiH);
  }
#if (DEBUG > 1) // default=0
  if (!err) { const AT_WC *feature=L"VerticallyCentreAOI";
    AT_IsImplemented(handle,feature,&b); 
    assert(b);
    AT_IsWritable(handle,feature,&b);
    assert(b);
    err = AT_GetBool(handle,feature,&b);
    if (!err) printf("verticalCenter=%d\n",b);
  }
#endif

  if (!err) { const AT_WC *feature=WSTR_AOI_L;
    err = AT_GetInt(handle,feature,&l);
#if (DEBUG > 1)
    if (!err) printf("AOIleft=%lld\n",l);
#endif
    assert(l == 1);
  }
  if (!err) { const AT_WC *feature=WSTR_AOI_T;
    err = AT_GetInt(handle,feature,&l);
#if (DEBUG > 1)
    if (!err) printf("AOItop=%lld\n",l);
#endif
    assert(l == 1);
  }
 
#if (DEBUG > 1)
  if (!err) { const AT_WC *feature=WSTR_BINH;
    AT_IsImplemented(handle,feature,&b); 
    assert(b);
    AT_IsWritable(handle,feature,&b);
    assert(b);
  }
  if (!err) { const AT_WC *feature=WSTR_BINV;
    AT_IsImplemented(handle,feature,&b); 
    assert(b);
    AT_IsWritable(handle,feature,&b);
    assert(b);
  }
#endif

#if 0 // =52083 ?what is this? 
  if (!err) { const AT_WC *feature=L"ExposedPixelHeight"; 
    err = AT_GetInt(handle,feature,&l);
    if (!err) printf("expPixelHeight=%lld\n",l);
  }
#endif

#if (DEBUG > 1) // default=1
  if (!err) { const AT_WC *feature=L"AccumulateCount";
    err = AT_GetInt(handle,feature,&l);
    if (!err) printf("accuCount=%lld\n",l);
    assert(l == 1);
  }
#endif

#if (DEBUG > 1) // default=1
  if (!err) { const AT_WC *feature=L"FrameCount";
    err = AT_GetInt(handle,feature,&l);
    if (!err) printf("frameCount=%lld\n",l);
    assert(l == 1);
  }
#endif
 
#if (DEBUG > 1) // default=Rolling (low read noise)
  if (!err) { const AT_WC *feature=L"ElectronicShutteringMode";
    err = get_enum_feature(handle,feature,buf,sizeof(buf));
    if (!err)  printf("shutter=%s\n",buf);
  }
#endif

#if 0 // not implemented at Zyla
  if (!err) { const AT_WC *feature=L"AOILayout";
    err = get_enum_feature(handle,feature,buf,sizeof(buf));
    if (!err)  printf("layout=%s\n",buf);
  }
#endif

  if (!err) { const AT_WC *feature=WSTR_BINNING;
    err = get_enum_feature(handle,feature,buf,sizeof(buf));
#if (DEBUG > 0)
    if (!err)  printf("binning=%s\n",buf);
#endif
  }

  if (!err) { const AT_WC *feature=L"FanSpeed"; /* Shec: always On */
    if (!err) err = get_enum_feature(handle,feature,buf,sizeof(buf));
#if (DEBUG > 0)
    if (!err) printf("fanSpeed=%s\n",buf);
#endif
    if (!err) err = AT_SetEnumString(handle,feature,L"On");
    if (!err) err = get_enum_feature(handle,feature,buf,sizeof(buf));
#if (DEBUG > 0)
    if (!err) printf("fanSpeed=%s\n",buf);
#endif
  }

#if (DEBUG > 1) // default=16 Bit
  if (!err) { const AT_WC *feature=L"BitDepth";
    err = get_enum_feature(handle,feature,buf,sizeof(buf));
    if (!err) printf("bitDepth=%s\n",buf);
  }
#endif

  if (!err) { const AT_WC *feature=L"PixelReadoutRate";
    if (!err) err = get_enum_feature(handle,feature,buf,sizeof(buf));
#if (DEBUG > 0)
    if (!err) printf("pixelRate=%s\n",buf);
#endif
#if 0 // TODO ?Shec 100 MHz / 270 MHz
    if (!err) err = AT_SetEnumString(handle,feature,L"100 MHz");
    //if (!err) err = AT_SetEnumString(handle,feature,L"270 MHz");
    if (!err) err = get_enum_feature(handle,feature,buf,sizeof(buf));
    if (!err) printf("pixelRate=%s\n",buf);
#endif
  }

#if (DEBUG > 1) // default=Centre Out Simultaneous
  if (!err) { char buf[128];
    err = get_enum_feature(handle,L"SensorReadoutMode",buf,sizeof(buf));
    if (!err) printf("readoutMode=%s\n",buf);
  }
#endif

#if (DEBUG > 1) // default=1
  if (!err) {
    err = AT_GetBool(handle,L"FastAOIFrameRateEnable",&b);
    if (!err) printf("fastFrames=%d\n",b);
  }
#endif

  if (!err) { const AT_WC *feature=L"Overlap"; // default=false
#if (DEBUG > 1)
    err = AT_GetBool(handle,feature,&b);
    if (!err) printf("overlap=%d\n",b);
#endif
    if (!err) err = AT_SetBool(handle,feature,AT_TRUE);
#if (DEBUG > 1)
    if (!err) err = AT_GetBool(handle,feature,&b);
    if (!err) printf("overlap=%d\n",b);
#endif
  }
 
#if (DEBUG > 1) // default=1
  if (!err) {
    err = AT_GetBool(handle,L"FullAOIControl",&b);
    if (!err) printf("fullAOIcontrol=%d\n",b);
  }
#endif
 
#if (DEBUG > 1) // default=0
  if (!err) {
    err = AT_GetBool(handle,L"ScanSpeedControlEnable",&b);
    if (!err) printf("scanSpeedControl=%d\n",b);
  }
#endif
 
#if (DEBUG > 1)
  double lineScanSpeed=1;
  if (!err) { double d;
    err = AT_GetFloat(handle,L"LineScanSpeed",&d);
    if (!err) printf("scanSpeed=%f (%f)\n",d,self->sensorH/d);
    if (!err) lineScanSpeed = d;
  }
#endif

#if (DEBUG > 1) // default=40 for USB3
  if (!err) { double d=0;
    err = AT_GetFloat(handle,L"MaxInterfaceTransferRate",&d);
    if (!err) printf("transferRate=%f\n",d);
  }
#endif

#if (DEBUG > 1) // default=6.5 
  if (!err) { double d=0;
    err = AT_GetFloat(handle,WSTR_PIXEL_H,&d);
    if (!err) printf("pixelHeight=%f\n",d);
  }
  if (!err) { double d=0;
    err = AT_GetFloat(handle,WSTR_PIXEL_W,&d);
    if (!err) printf("pixelWidth=%f\n",d);
  }
#endif

  if (!err) { double d=0;
    err = AT_GetFloat(handle,WSTR_RTIME,&d);
#if (DEBUG > 1)
    if (!err) printf("readoutTime=%f (%f)\n",d,
                     self->sensorH/(2.0*lineScanSpeed));
#endif
    self->readoutTime = d; 
  }

#if (DEBUG > 1)
  if (!err) { double d=0;
    err = AT_GetFloat(handle,L"RowReadTime",&d);
    if (!err) printf("rowTime=%f (%f)\n",d,d*self->sensorH);
  }
#endif

#if (DEBUG > 1) // default=1
  if (!err) { const AT_WC *feature=L"SpuriousNoiseFilter";
    err = AT_GetBool(handle,feature,&b); 
    if (!err) printf("noiseFilter=%d\n",b);
#if 0 // disabled -- enable for testing
    if (!err) err = AT_SetBool(handle,feature,AT_FALSE);
    err = AT_GetBool(handle,feature,&b); 
    if (!err) printf("noiseFilter=%d\n",b);
#endif
  }
#endif
 
#if (DEBUG > 1) // default=1
  if (!err) { const AT_WC *feature=L"StaticBlemishCorrection";
    err = AT_GetBool(handle,feature,&b); 
    if (!err) printf("blemishCorrection=%d\n",b); 
#if 0 // disabled -- enable for testing
    if (!err) err = AT_SetBool(handle,feature,AT_FALSE);
    err = AT_GetBool(handle,feature,&b); 
    if (!err) printf("blemishCorrection=%d\n",b); 
#endif
  }
#endif
 
#if (DEBUG > 1)
  if (!err) {
    if (!err) err = AT_GetInt(handle,L"TimestampClock",&l);
    if (!err) printf("clock=%lld\n",l);
    if (!err) err = AT_GetInt(handle,L"TimestampClockFrequency",&l);
    if (!err) printf("freq=%lld\n",l);
  }
#endif

#if (DEBUG > 1) // default=Internal
  if (!err) {
    err = get_enum_feature(handle,L"TriggerMode",buf,sizeof(buf));
    if (!err) printf("triggerMode=%s\n",buf);
  }
#endif

  if (!err) { const AT_WC *feature=L"SimplePreAmpGainControl";
    AT_IsImplemented(handle,feature,&b);
    assert(b);
    err = AT_SetEnumString(handle,feature,
              L"16-bit (low noise & high well capacity)");
  }
#if (DEBUG > 1) // default=Mono16 after set to 16-bit
  if (!err) { const AT_WC *feature=L"PixelEncoding";
    err = get_enum_feature(handle,feature,buf,sizeof(buf));
    if (!err) printf("pixelEncoding=%s\n",buf);
  }
#endif
#if (DEBUG > 1) // default=Mono16 after set to 16-bit 
  if (!err) { const AT_WC *feature=L"PixelEncoding";
    AT_IsImplemented(handle,feature,&b);
    assert(b);
    AT_IsWritable(handle,feature,&b);
    assert(b);
    err = AT_SetEnumString(handle,feature,L"Mono16");
  }
#endif

  if (!err) err = andor_tempcon(self,1);
  if (!err) err = andor_exptime(self,0.5);
  if (!err) self->initialized = 1;

  self->err = err;

  return err;  
}

/* ---------------------------------------------------------------- */

int andor_close(AndorStruct* self)
{
  int err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%s(%p) self.err=%d\n",__FILE__,PREFUN,self,self->err);
#endif
  assert(self);

  pthread_mutex_lock(&self->ioLock);

  if (self->handle >= 0) {
    if (!self->err) err = AT_Close(self->handle); /* Note: crashes */
  }
  self->handle = -1;
  self->initialized = 0;
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_free(AndorStruct* self)
{
  int i;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%s(%p)\n",__FILE__,PREFUN,self);
#endif
  assert(self);

  pthread_mutex_destroy(&self->ioLock);
  pthread_mutex_destroy(&self->frameLock);
  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *frame = &self->frames[i];
    if (frame->data) free((void*)frame->data);
  }
  free((void*)self);

  pthread_mutex_lock(&andor_mutex);
  if (andor_initialized > 0)  {
    andor_initialized -= 1;
    if (andor_initialized == 0) AT_FinaliseLibrary();
  }
  pthread_mutex_unlock(&andor_mutex);

  return andor_initialized;
}

/* ---------------------------------------------------------------- */

int andor_aoi(AndorStruct* self,int dim,int bin)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%d,%d)\n",PREFUN,self,dim,bin);
#endif
  assert(self);
  assert(self->handle >= 0);
  assert(self->tid == 0);

  int err = andor_binningXY(self,bin);

  pthread_mutex_lock(&self->ioLock);

  if (!err) err = AT_SetInt(self->handle,WSTR_AOI_W,dim/bin);
  assert(self->sensorW >= dim);
  if (!err) err = AT_SetInt(self->handle,WSTR_AOI_L,1+(self->sensorW-dim)/2);
  if (!err) err = AT_SetInt(self->handle,WSTR_AOI_H,dim/bin);
  assert(self->sensorH >= dim);
  if (!err) err = AT_SetInt(self->handle,WSTR_AOI_T,1+(self->sensorH-dim)/2);

  if (!err) { AT_64 w,h,l,t,x,y;
    if (!err) err = AT_GetInt(self->handle,WSTR_AOI_W,&w);
    if (!err) self->aoiW = w;
    if (!err) err = AT_GetInt(self->handle,WSTR_AOI_H,&h);
    if (!err) self->aoiH = h;
    AT_GetInt(self->handle,WSTR_AOI_L,&l);
    AT_GetInt(self->handle,WSTR_AOI_T,&t);
    AT_GetInt(self->handle,WSTR_BINH,&x);
    AT_GetInt(self->handle,WSTR_BINV,&y);
#if (DEBUG > 1)
    if (!err) printf("AOI: %lld,%lld,%lld,%lld (%lld,%lld)\n",w,h,l,t,x,y);
#endif
  }

  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_temperature(AndorStruct *self)
{
  double value;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%s(%p)\n",__FILE__,PREFUN,self);
#endif
  assert(self->handle >= 0);
  if (self->handle < 0) return E_NOTINIT;

  pthread_mutex_lock(&self->ioLock);

  int err = AT_GetFloat(self->handle,WSTR_STEMP,&value);
  if (!err) self->tempSensor = (float)value;
  if (!err) { char buf[128];
    err = get_enum_feature(self->handle,WSTR_TSTAT,buf,sizeof(buf));
    assert(strlen(buf) < sizeof(self->tempStatus));
    if (!err) strcpy(self->tempStatus,buf);
  }
  if (err) self->err = err;

  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_tempcon(AndorStruct* self,int value)
{
  pthread_mutex_lock(&self->ioLock);

  int err = AT_SetBool(self->handle,WSTR_COOLING,
                       (value) ? AT_TRUE : AT_FALSE);

  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_exptime(AndorStruct* self,double exptime)
{
  int err=0;
  const AT_WC *feature=WSTR_EXPTIME;

  pthread_mutex_lock(&self->ioLock);
  
#if (DEBUG > 1)
  double min=0,max=0;
  if (!err) err = AT_GetFloatMin(self->handle,feature,&min);
  if (!err) err = AT_GetFloatMax(self->handle,feature,&max);
  if (!err) printf("min=%.3f, max=%.3f\n",min,max);
#endif

  err = AT_SetFloat(self->handle,feature,exptime);
  if (!err) {
    self->expTime = exptime; 
    err = AT_GetFloat(self->handle,feature,&exptime);
    if (!err) {
      if (fabs(self->expTime - exptime) > 0.001) {
        self->expTime = exptime; 
        err = E_andor_setpar;
      }
    }
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_binningX(AndorStruct* self,int value)
{
  const AT_WC *feature=WSTR_BINH;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  pthread_mutex_lock(&self->ioLock);

  int err = AT_SetInt(self->handle,feature,value);
  if (!err) { AT_64 v;
    self->binH = value; 
    err = AT_GetInt(self->handle,feature,&v);
    if (!err) {
      if (self->binH != v) {
        self->binH = (int)v; 
        err = E_andor_setpar;
      }
    }
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_binningY(AndorStruct* self,int value)
{
  const AT_WC *feature=WSTR_BINV;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  pthread_mutex_lock(&self->ioLock);

  int err = AT_SetInt(self->handle,feature,value);
  if (!err) { AT_64 v;
    self->binV = value; 
    err = AT_GetInt(self->handle,feature,&v);
    if (!err) {
      if (self->binV != v) {
        self->binV = (int)v; 
        err = E_andor_setpar;
      }
    }
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

int andor_binningXY(AndorStruct* self,int value)
{
  const AT_WC *feature=WSTR_BINNING;
  AT_WC wbuf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,value);
#endif

  pthread_mutex_lock(&self->ioLock);

  swprintf(wbuf,128,L"%dx%d",value,value);
  int err = AT_SetEnumString(self->handle,feature,wbuf);
  if (!err) { char buf[128],baf[128];
    err = get_enum_feature(self->handle,feature,buf,sizeof(buf));
    if (!err) { 
      wcstombs(baf,wbuf,sizeof(baf));
      // printf("binning=%s %s\n",buf,baf);
      if (strcmp(buf,baf)) err = E_andor_setpar;
      if (!err) { AT_64 w=0,h=0,l=0,t=0;
        self->binH = self->binV = value;
        if (!err) err = AT_GetInt(self->handle,WSTR_AOI_W,&w);
        if (!err) self->aoiW = w;
        if (!err) err = AT_GetInt(self->handle,WSTR_AOI_H,&h);
        if (!err) self->aoiH = h;
        AT_GetInt(self->handle,WSTR_AOI_L,&l);
        AT_GetInt(self->handle,WSTR_AOI_T,&t);
        // printf("AOI: %lld,%lld,%lld,%lld\n",w,h,l,t);

      }
    }
  }
  pthread_mutex_unlock(&self->ioLock);

  return err;
}

/* ---------------------------------------------------------------- */

static void* run_cycle(void* param)
{
  int err=0;
  AndorStruct *self=(AndorStruct*)param;
  u_char *puc_returnBuf; 
  int bufSize;
  double t1,t2;
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%s(%p)\n",__FILE__,PREFUN,param);
#endif
  assert(self->handle >= 0);
  message(self,PREFUN,MSG_FLUSH);

  pthread_mutex_lock(&self->ioLock);

  if (!err) {
    err = AT_SetEnumString(self->handle,L"CycleMode",L"Continuous");
    // printf("err(mode)=%d\n",err);
  }
  if (!err) {
    err = AT_Command(self->handle,WSTR_START);
    // printf("err(start)=%d\n",err);
  }
  t1 = walltime(0);

  pthread_mutex_unlock(&self->ioLock);

  self->fps = 0;
  if (!err) {
    self->err = 0;
    while (!self->stop_flag) {
      msleep(5); /* IDEA query temperature here ? */
      pthread_mutex_lock(&self->ioLock);
      err = AT_WaitBuffer(self->handle,&puc_returnBuf,&bufSize,2000);
      pthread_mutex_unlock(&self->ioLock);
      if (self->stop_flag) break;
      if (err == AT_ERR_TIMEDOUT) { double dt = walltime(0)-t1;
        if (dt > 2+MAX_EXPTIME) {
          sprintf(buf,"%s: err=%d",PREFUN,err);
          message(self,buf,MSG_FLUSH);
          t1 = walltime(0);
        }
        continue;
      }
      if (err) break;                  /* other error -- terminate */
      t2 = walltime(0);
      self->fps = 0.5*self->fps + 0.5/(t2-t1);
      t1 = t2;
      self->seqNumber += 1;
      AndorFrame *frame = andor_frame4writing(self,self->seqNumber);
      if (frame) {
        frame->seqNumber = self->seqNumber;
        frame->timeStamp = t2 - self->expTime - self->readoutTime;
        memcpy(frame->data,puc_returnBuf,bufSize);
        andor_frame_release(self,frame);
      }
      pthread_mutex_lock(&self->ioLock);
      err = AT_QueueBuffer(self->handle,puc_returnBuf,bufSize);
      pthread_mutex_unlock(&self->ioLock);
      if (err) {  
        sprintf(buf,"%s: PANIC -- buffer queue failed, err=%d",PREFUN,err);
        message(self,buf,MSG_FILE);
      }
    } /* endwhile (!stop_flag) */
  } /* endif(!err) */
  free((void*)self->data_1); self->data_1 = NULL;
  free((void*)self->data_2); self->data_2 = NULL;

  if (err == AT_ERR_COMM) self->err = err;
  sprintf(buf,"%s/%s: done, err=%d",__FILE__,PREFUN,err);
#if (DEBUG > 0)
  fprintf(stderr,"%s\n",buf);
#endif
  if (err) message(self,buf,MSG_FLUSH);

  return (void*)(long)err;
}

/* ---------------------------------------------------------------- */

int andor_cycle_start(AndorStruct* self)
{
  int err=0;

  message(NULL,PREFUN,MSG_FLUSH);

  pthread_mutex_lock(&self->ioLock);

  size_t size = self->aoiW * self->aoiH * sizeof(u_short);

  self->data_1 = (u_char*)malloc(size);
  assert(((u_long)self->data_1 & 0x07) == 0);
  if (!err) err = AT_QueueBuffer(self->handle,self->data_1,(int)size);

  self->data_2 = (u_char*)malloc(size);
  assert(((u_long)self->data_2 & 0x07) == 0);
  if (!err) err = AT_QueueBuffer(self->handle,self->data_2,(int)size);

  pthread_mutex_unlock(&self->ioLock);

  if (!err) { int i;
    for (i=0; i<ANDOR_NBUFS; i++) {
      AndorFrame *frame = &self->frames[i];    
      frame->timeStamp = 0;
      frame->seqNumber = 0;
      assert(!frame->data);
      frame->data = (u_short*)malloc(size);
      assert(((u_long)frame->data & 0x07) == 0);
      frame->w = self->aoiW;
      frame->h = self->aoiH;
      frame->wlock = 0;
      frame->rlock = 0;
    }
  }

  if (!err) {
    self->stop_flag = 0;
    thread_create(run_cycle,(void*)self,&self->tid);
  } else {
    self->tid = 0;
  }

  return err;
}

/* ---------------------------------------------------------------- */

int andor_cycle_stop(AndorStruct *self)
{
  int  err=0,i;
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%s(%p)\n",__FILE__,PREFUN,self);
#endif
  message(NULL,PREFUN,MSG_FLUSH);

  pthread_mutex_lock(&self->ioLock);
  if (!err) {
    err = AT_Command(self->handle,L"Acquisition Stop");
    sprintf(buf,"%s: err(stop)=%d",PREFUN,err);
    message(self,buf,MSG_FILE);
  }
  pthread_mutex_unlock(&self->ioLock);

  self->stop_flag = 1;
  pthread_join(self->tid,NULL); self->tid = 0;

  pthread_mutex_lock(&self->ioLock);
  if (!err) {
    err = AT_Flush(self->handle);
    sprintf(buf,"%s: err(flush)=%d",PREFUN,err);
    message(self,buf,MSG_FILE);
  }
  pthread_mutex_unlock(&self->ioLock);

  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *f = &self->frames[i]; 
    while (f->wlock != 0) { 
      msleep(50); fprintf(stderr,"%s: f=%d, wlock=%d\n",PREFUN,i,f->wlock); 
    }
    while (f->rlock != 0) { 
      msleep(50); fprintf(stderr,"%s: f=%d, rlock=%d\n",PREFUN,i,f->rlock); 
    } 
    if (f->data) { free((void*)f->data); f->data = NULL; }
  }

  return err;
}

/* ---------------------------------------------------------------- */

AndorFrame* andor_frame4writing(AndorStruct* self,u_int s)
{
  int i;
  AndorFrame *frame=NULL;

  pthread_mutex_lock(&self->frameLock);
  
  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *f = &self->frames[i];    
    assert(f->data);
    assert(f->wlock == 0);             // no frame locked for writing
    if (f->rlock == 0) {               // not locked for reading
      if (f->seqNumber < s) {          // oldest frame
        frame = f;
        s = frame->seqNumber;
      }
    }
  }
  if (!frame) { char buf[128];
    sprintf(buf,"PANIC: no frame available for writing");
    message(self,buf,MSG_FILE);
  } else {
    frame->wlock = 1;
  }
  pthread_mutex_unlock(&self->frameLock);
        
  return frame;
}

/* ---------------------------------------------------------------- */

AndorFrame* andor_frame4reading(AndorStruct* self,u_int s)
{
  int i;
  AndorFrame *frame=NULL;

  if (self->stop_flag) return NULL;

  pthread_mutex_lock(&self->frameLock);
  
  for (i=0; i<ANDOR_NBUFS; i++) {
    AndorFrame *f = &self->frames[i];    
    assert(f->data);
    if (f->wlock == 0) {               // not locked for writing
      if (f->seqNumber > s) {          // newest frame
        frame = &self->frames[i];
        s = frame->seqNumber;
      }
    }
  }
  if (frame) {
    frame->rlock += 1;
  }
  pthread_mutex_unlock(&self->frameLock);
        
  return frame;
}


/* ---------------------------------------------------------------- */

void andor_frame_release(AndorStruct* self,AndorFrame *frame)
{
  pthread_mutex_lock(&self->frameLock);

  if (frame->wlock) {                  /* is locked for writing */
    assert(frame->rlock == 0);
    assert(frame->wlock == 1);
    frame->wlock = 0;
  } else {                             /* is locked for reading */
    assert(frame->rlock >  0);
    frame->rlock -= 1;
    assert(frame->rlock >= 0);
  }
  
  pthread_mutex_unlock(&self->frameLock);
}

/* ---------------------------------------------------------------- */

int andor_sim(AndorStruct* self,const char* command,char* res)
{
  int err=0;
  char cmd[128]="",par[128]="";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%s,%p): %d\n",PREFUN,self,command,res,self->handle);
#endif

  if (strlen(command) >= sizeof(cmd)) return -1;
  sscanf(command,"%s %s",cmd,par); 

#ifdef SIM_ONLY
  assert((self->handle >= 0) && (self->handle < MAX_NANDORS));
  if (!strncasecmp(cmd,"simfwhm",5)) {
    if (*par) sim_sigma[self->handle] = fmax(1.0,atof(par)/2.35482);
    if (res) sprintf(res,"%f",2.35482*sim_sigma[self->handle]);
  } else 
  if (!strncasecmp(cmd,"simmode",5)) {
    if (*par) sim_mode[self->handle] = atoi(par);
    if (res) sprintf(res,"%d",sim_mode[self->handle]);
  } else
  if (!strncasecmp(cmd,"simradius",5)) {
    if (*par) sim_radius[self->handle] = fmax(0.0,atof(par));
    if (res) sprintf(res,"%f",sim_radius[self->handle]);
  } else
  if (!strncasecmp(cmd,"simflux",5)) {
    if (*par) sim_flux[self->handle] = 1.0e5*pow(5.0,atof(par));
    if (res) sprintf(res,"%f",log(sim_flux[self->handle]/1.0e5)/log(5.0));
  } else {
    if (res) sprintf(res,"-Eunknown");
    err = -2;
  }
#endif

  return err;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

