/* ----------------------------------------------------------------
 *
 * andorserver.c  
 * 
 * Project: Andor Camera software (OCIW, Pasadena, CA)
 *
 * v0.200  2018-05-09  Christoph C. Birk, birk@carnegiescience.edu
 *
 * NOTE: systemctl stop firewalld
 *       systemctl disable firewalld
 *
 * VSC-04939 (LCO)
 * VSC-07838,VSC-07839,VSC-07840 (SBS)
 *
 * ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           1              /* debug level */
#endif

#define TIME_TEST       1

/* ------------------------------------------------------- */

#define _REENTRANT

#include <stdlib.h>                    /* atoi(),exit() */
#include <stdio.h>                     /* sprintf() */
#include <string.h>                    /* strcpy(),memset(),memcpy() */
#include <assert.h>
#include <errno.h>

#include <sys/reboot.h>                /* requires 'andorserver' */
#include <linux/reboot.h>              /* to run as 'root' */

#if (TIME_TEST > 0)
#include <limits.h> 
#include <sys/times.h>
#include <unistd.h>
#endif

#include "andor.h"
#include "andorUSB.h" 
#include "utils.h"                     /* generic utilities */
#include "random.h"
#include "ptlib.h"                     /* POSIX threads lib */
#include "tcpip.h"                     /* TCPIP stuff */

/* DEFINEs -------------------------------------------------------- */

#define P_TITLE         "AndorServer"

#define PREFUN          __func__

/* STATICs -------------------------------------------------------- */

enum server_states { STATE_OFF=0,
                     STATE_CONNECTED,
                     STATE_READY,
                     STATE_RUNNING };

static char       logfile[512];
static int        offtime=0,baseD=2048,baseB=1;
static int        tcpDebug=0;
static u_int      cookie=0,tcpNumber=0;

static char       *data_buffer=NULL;

static int          andor_id=0;        /* {0,1,2} */
static time_t       startup_time;
static AndorStruct  *andor=NULL;

static volatile int state=STATE_OFF;

#ifdef SIM_ONLY
static char         fake_serialNumber[32]="sim0001";
#endif

/* function prototype(s) ------------------------------------------ */

static int     do_state_change   (int);
static void*   run_housekeeping  (void*);
static void*   run_tcpip         (void*);

/* --- M A I N ---------------------------------------------------- */

int main(int argc,char **argv)
{
  int  i,connectAtStartup=0;
  char buf[128],buf2[128],buffer[1024];
#if (DEBUG > 0)
  fprintf(stderr,"%s-%s\n",P_TITLE,P_VERSION);
#endif

  /* here comes the code ------------------------------------------ */ 

  startup_time = cor_time(0);

  { extern char *optarg;               /* parse command line */  
    extern int opterr,optopt; opterr=0;
    while ((i=getopt(argc,argv,"di:kos:w:")) != EOF) {
      switch (i) {
      case 'd':                        /* debug: allow mult. connections */
        tcpDebug = 1;
        break;
      case 'i':                        /* Andor driver 'ID' {0,1,2} */
        andor_id = atoi(optarg);
#ifdef SIM_ONLY
        sprintf(fake_serialNumber,"sim%04d",1+andor_id);
#endif
        break;
      case 'k':                        /* kill old instance */
        TCPIP_SingleCommand("localhost",SERVER_PORT+andor_id,"quit\n");
        sleep(1);
        break;
      case 'o':                        /* open USB connection */
        connectAtStartup = 1;
        break;
      case 's':                        /* fake serial number */
#ifdef SIM_ONLY
        strcpy(fake_serialNumber,optarg);
#endif
        break;
      case 'w':                        /* wait */
        sleep(atoi(optarg));
        break;
      case '?':
        fprintf(stderr,"%s: option '-%c' unknown or parameter missing\n",
                P_TITLE,(char)optopt);
        break;
      }
    } /* endwhile(getopt) */
  }

  InitRandom(0,0,0);                   /* initialize random# */
  cookie = URandom();

  /* -------------------------------------------------------------- */

  sprintf(buffer,"andorserver%d_",1+andor_id); 
  log_path(logfile,"ANDORLOG",buffer);
  sprintf(buffer,"andorserver%d.log",1+andor_id); // todo rotate logfile ?
  lnk_logfile(logfile,buffer);
  sprintf(buffer,"%s-v%s: UT= %s LT= %s",P_TITLE,P_VERSION,
          get_ut_timestr(buf,0),get_local_timestr(buf2,0));
  message(NULL,buffer,MSG_FLUSH);

#ifndef SIM_ONLY
  for (i=8; i>=1; i--) {               /* rotate atdebug.log */
    sprintf(buf,"mv /tmp/atdebug%d.log /tmp/atdebug%d.log",i,i+1);
    system(buf); // printf("%s\n",buf);
  }
  system("mv /tmp/atdebug.log /tmp/atdebug1.log");
#endif
  
  andor = andor_create(andor_id);
  assert(andor);
  if (connectAtStartup) do_state_change(STATE_CONNECTED);

  /* -------------------------------------------------------------- */

  thread_detach(run_housekeeping,NULL);
  run_tcpip(NULL);                     /* blocking this thread */

  /* -------------------------------------------------------------- */

  assert(andor);
  if (andor->tid) msleep(2500);
  andor_close(andor);
  andor_free(andor);

#if (TIME_TEST > 0)                    /* print cpu-time */
  { struct tms tmsbuf;
    (void)times(&tmsbuf);
    sprintf(buffer,"%s(%d): total-cpu: %5.2f seconds",P_TITLE,
      (int)getpid(),
      (float)(tmsbuf.tms_stime)/(float)sysconf(_SC_CLK_TCK) +
      (float)(tmsbuf.tms_utime)/(float)sysconf(_SC_CLK_TCK));
    message(NULL,buffer,MSG_FLUSH); 
  }
#endif
 
  return 0;
}

/* ---------------------------------------------------------------- */

static int do_state_change(int target)
{
  int err=0,wait=2000;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d): state=%d\n",PREFUN,target,state);
#endif

  switch (state) {
  case STATE_OFF:
    assert(andor->handle < 0);
    assert(andor->tid == 0);
    if (target > state) {              /* --> Connected */
      err = andor_open(andor);         /* get serial,model, etc */
      if (!err) {
        baseD = andor->sensorW;
        baseB = 1;
        state++;
      }
    }
    break;
  case STATE_CONNECTED:
    assert(andor->handle >= 0);
    assert(andor->tid == 0);
    if (target < state) {              /* --> Off */
      err = andor_close(andor);
      if (!err) state--;
    } else 
    if (target > state) {              /* --> Ready */
      if (!err) err = andor_init(andor);
      if (!err) err = andor_aoi(andor,baseD,baseB);
      if (data_buffer) free((void*)data_buffer);
      data_buffer = (char*)calloc(andor->aoiW*andor->aoiH,sizeof(short));
      if (!err) state++;
    }
    break;
  case STATE_READY:
    assert(andor->handle >= 0);
    assert(andor->tid == 0);
    if (target < state) {              /* --> Connected */
      state--;
    } else 
    if (target > state) {              /* --> Running */
      message(NULL,"starting acquisition",MSG_FLUSH);
      err = andor_cycle_start(andor);
      if (!err) {
        while (wait > 0) {
          msleep(350); wait -= 350; if (andor->tid) break;
        }
        if (andor->tid) state++;
        else            err = E_RUNNING;
      }
    }
    break;
  case STATE_RUNNING:
    assert(andor->handle >= 0);
    if (target < state) {              /* --> Ready */
      if (andor->tid) {                /* acq. is running */
        message(NULL,"stopping acquisition",MSG_FLUSH);
        (void)andor_cycle_stop(andor);
        assert(!andor->tid);
        state--;
      }
    }     
    break;
  default:
    assert(0);
    err = 234; // unknown state
    break;
  }
  if (!err && (state != target)) {
    err = do_state_change(target);
  }

  return err;
}

/* ---------------------------------------------------------------------- */

void message(const void* param,const char* text,int flags)
{
  static FILE* fp=NULL;
  char tstr[32];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d,%p)\n",PREFUN,text,flags,file);
#endif
  assert(text);

  get_ut_datestr(tstr,cor_time(offtime));
#ifndef NDEBUG
  fprintf(stderr,"%s: %s\n",tstr,text);
#endif

  if (flags & MSG_FILE) {              /* write to file */
    if (!fp) fp = fopen(logfile,"a");
    if (fp) { char lev[128];
      if      (flags & MSG_RED)    strcpy(lev,"ERROR");
      else if (flags & MSG_YELLOW) strcpy(lev,"WARNING");
      else                         strcpy(lev,"-");
      fprintf(fp,"%s - %s: %s\n",tstr,lev,text);
      if (flags & MSG_FLUSH) fflush(fp);
    } else {
      fprintf(stderr,"%s:%s: opening %s failed\n",__FILE__,PREFUN,logfile);
    }
  }
}

/* ---------------------------------------------------------------- */

static void* run_housekeeping(void* param)
{
  int    err,ecnt=0,done=0;
  time_t temp_next=0,temp_interval=2;
#if (DEBUG > 0)
  fprintf(stderr,"%s: %s(%p)\n",__FILE__,PREFUN,param);
#endif

  while (!done) {
    sleep(1);
    if (state >= STATE_CONNECTED) {
      if (cor_time(0) >= temp_next) {
        temp_next = cor_time(0)+temp_interval;
        temp_interval = imin(temp_interval+1,60);
        err = andor_temperature(andor);
        if (err) { char buf[128];      /* 17=AT_ERR_COMM */
          sprintf(buf,"%s(): err=%d (%d)",PREFUN,err,++ecnt);
          message(NULL,buf,MSG_ERROR); 
          strcpy(andor->tempStatus,"stale");
          if (ecnt >= 3) {             /* give up */
            strcpy(andor->serialNumber,"lost");
            do_state_change(STATE_OFF);
          }
        } else { ecnt=0; } 
      } /* endif(_next) */
    } else { temp_interval=2; ecnt=0; }
  }
  return (void*)0;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

static int handle_command(const char* command,char* answer,size_t buflen)
{
  int  r=0;
  long err=0;
  char cmd[128]="",par1[128]="",par2[128]="",buf[128];

  strcpy(answer,"-Einvalid command\n");
  if (strlen(command) >= sizeof(cmd)) return 0;
  int n = sscanf(command,"%s %s %s",cmd,par1,par2);
  if (n < 1) return 0;

  strcpy(answer,"OK");                 /* default response */
  if (!strcasecmp(cmd,"next")) {       /* send next frame */
    if (state != STATE_RUNNING) {
      err = E_RUNNING;
    } else { AndorFrame *frame=NULL; double timeout;
      timeout = (*par1) ? atof(par1) : 1.0;
      double t1 = walltime(0) + timeout;
      do {                             /* wait for frame */
        frame = andor_frame4reading(andor,tcpNumber);
        if (frame) break;
        msleep(10);
      } while(walltime(0) < t1);
      if (frame) {
        tcpNumber = frame->seqNumber;
        size_t size = andor->aoiW*andor->aoiH*sizeof(u_short);
        memcpy(data_buffer,frame->data,size);
        andor_frame_release(andor,frame);
        assert(andor->aoiW == andor->aoiH);
        sprintf(answer,"%u %d %.2f %.2f",
                frame->seqNumber,andor->aoiW,
                frame->timeStamp-offtime, /* exptime,readoutTime @ USB */
                andor->tempSensor);
        r = 1;
      } else {
        strcpy(answer,"-Enodata");
      }
    }
  } else
  if (!strncasecmp(cmd,"version",3)) { /* server version */
    sprintf(answer,"%s %u %ld",P_VERSION,cookie,startup_time);
  } else
  if (!strncasecmp(cmd,"serial",3)) {  /* Andor serial number */
    if (state < STATE_CONNECTED) {
      err = do_state_change(STATE_CONNECTED);
    }
    if (!err) {
#ifdef SIM_ONLY
      strcpy(andor->serialNumber,fake_serialNumber);
#endif
      strcpy(answer,andor->serialNumber);
    }
  } else 
  if (!strncasecmp(cmd,"setup",3)) {   /* setup geometry */
    err = do_state_change(STATE_CONNECTED);
    if (*par1) baseD = atoi(par1);
    if (*par2) baseB = atoi(par2);
    err = do_state_change(STATE_READY);
  } else 
  if (!strcasecmp(cmd,"start")) {      /* start data acquisition */
    if (state < STATE_RUNNING) {
      err = do_state_change(STATE_RUNNING);
    }
  } else 
  if (!strcasecmp(cmd,"stop")) {       /* stop data acquisition */
    if (state >= STATE_RUNNING) {
      err = do_state_change(STATE_READY);
    }
  } else 
  if (!strcasecmp(cmd,"close")) {      /* close connection to Andor */
    err = do_state_change(STATE_OFF);
  } else 
  if (!strcasecmp(cmd,"state")) {      /* server state */
    assert(andor->aoiW == andor->aoiH);
    assert(andor->binH == andor->binV);
    sprintf(answer,"%d %d %d %u %u",state,
            andor->aoiW,andor->binH,
            andor->seqNumber,tcpNumber);
  } else 
  if (!strncasecmp(cmd,"temperature",5)) {
    if (state == STATE_CONNECTED) err = andor_temperature(andor);
    if (state < STATE_RUNNING) strcpy(andor->tempStatus,"stale");
    sprintf(answer,"%.2f %s",andor->tempSensor,andor->tempStatus);
  } else
  if (!strncasecmp(cmd,"tempcon",5)) { int value; int oldState=state;
    if      (!strcasecmp(par1,"off")) value = 0;
    else if (!strcasecmp(par1,"on"))  value = 1;
    else                              value = atoi(par1);
    if (state != STATE_READY) {
      err = do_state_change(STATE_READY);
    }
    if (!err) err = andor_tempcon(andor,value);
    if (!err) { 
      if (oldState == STATE_RUNNING) err = do_state_change(oldState);
    }
  } else
  if (!strncasecmp(cmd,"exptime",3)) {
    if (n > 1) {
      float t = (float)fmin(MAX_EXPTIME,fmax(0.01,atof(par1)));
      err = andor_exptime(andor,t);
    }
    if (!err) sprintf(answer,"%.3f",andor->expTime);
  } else
  if (!strcasecmp(cmd,"offtime")) {
    if (*par1) {
      offtime = cor_time(0) - atol(par1);
    }
    sprintf(buf,"offtime= %d",offtime);
    message(NULL,buf,MSG_FILE);
    sprintf(answer,"%d",offtime);
  } else
  if (!strncasecmp(cmd,"sim",3)) {
    err = andor_sim(andor,command,buf);
    if (!err) strcpy(answer,buf);
  } else
  if (!strncasecmp(cmd,"quit",4)) {    /* terminate */
    (void)do_state_change(STATE_OFF);
    message(NULL,cmd,MSG_FLUSH);
    exit(0);
  } else 
  if (!strncasecmp(cmd,"reboot",4)) {  /* reboot */
    do_state_change(STATE_OFF);
    message(NULL,cmd,MSG_FLUSH);
    r = 2;                             /* terminate and reboot */
  } else
  if (!strncasecmp(cmd,"restart",4)) { /* restart myself */
    do_state_change(STATE_OFF);
    message(NULL,cmd,MSG_FLUSH);
    r = 3;                             /* terminate and restart */
  } else
  if (!strncasecmp(cmd,"shutdown",4) || !strncasecmp(cmd,"poweroff",4)) {
    do_state_change(STATE_OFF);        /* power down MinnowBoard */
    message(NULL,cmd,MSG_FLUSH);
    r = 4;
  } else { /* bug-fix NEW 2019-01-09 */
    strcpy(answer,"-Eunknown command");
  }
  if (err) sprintf(answer,"-Eerr=%ld",err);
#if (DEBUG > 1)
  sprintf(cmd,"%s: send '%s'",PREFUN,answer);
  message(NULL,cmd,MSG_FILE);
#endif
  strcat(answer,"\n");

  return r;
}

/* --- */

static int receive_string(int sock,char* string,size_t buflen)
{
  int     i=0;
  ssize_t r;

  do {
    r = recv(sock,&string[i],1,0);
    if (r != 1) break;
    if (string[i] == '\n') break;      /* <LF> */
    if (string[i] == '\r') break;      /* <CR> */
    i++;
  } while (i < buflen-1);
  string[i] = '\0';                    /* overwrite term-char */

  return r;
}

/* --- */

typedef struct {
  char host[128];
  int  port,msgsock;
} Connection;
  
/* --- */

static void* run_connection(void* param)
{
  Connection *c = (Connection*)param;
  int  rval,r;
  long done=0;
  char cmd[128],buf[256];

  do {
    rval = receive_string(c->msgsock,cmd,sizeof(cmd));
    if (rval > 0) {
#if (DEBUG > 1)
      sprintf(buf,"%s(): received '%s'",PREFUN,cmd);
      message(NULL,buf,MSG_FILE);
#endif
      r =  handle_command(cmd,buf,sizeof(buf));
      send(c->msgsock,buf,strlen(buf),MSG_NOSIGNAL);
      if (r == 1) {                    /* send frame */
        size_t size = andor->aoiW*andor->aoiH*sizeof(u_short);
        send(c->msgsock,data_buffer,size,MSG_NOSIGNAL);
      } else
      if (r == 2) {                    /* reboot MinnowBoard */
        sync(); sleep(1);              /* sync disks */
#ifndef SIM_ONLY
        reboot(LINUX_REBOOT_CMD_RESTART); /* Note: requires 'root' */
#endif
        exit(0);
      } else
      if (r == 3) {                    /* restart TODO remove ? */
        sprintf(buf,"nohup %s/andorserver -w 10 -i %d &",getenv("HOME"),andor_id);
        system(buf);
        exit(0);
      } else 
      if (r == 4) {                    /* power down MinnowBoard */
        sync(); sleep(1);              /* sync disks */
#ifndef SIM_ONLY
        reboot(LINUX_REBOOT_CMD_POWER_OFF); /* Note: requires 'root' */
#endif
        exit(0);
      }
    } else {
      sprintf(buf,"%s(%s): hangup",PREFUN,c->host);
      message(NULL,buf,MSG_FLUSH);
    }
  } while (rval > 0);                  /* while there's something */
  (void)close(c->msgsock);
  free((void*)c);

  return (void*)done;
}

/* --- */

static void* run_tcpip(void* param)
{
  int    port=SERVER_PORT+andor_id;
  const int done=0;
  int    err,msgsock;
  char   buf[256],host[128];
  struct sockaddr sadr;
  socklen_t size=sizeof(sadr);
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,port);
#endif

  int sock = TCPIP_CreateServerSocket(port,&err);
  if (err) {
    sprintf(buf,"%s(%d): err=%d",PREFUN,port,err);
    message(NULL,buf,MSG_FLUSH);
    return (void*)(long)err;
  }

  while (!done) {
    // NOTE: if someone tries to connect while we are busy
    //       the requests get queued -- IDEA flush ?
    msgsock = accept(sock,&sadr,&size);   /* accept connection */
#ifdef SO_NOSIGPIPE                    /* 2017-07-05 */
    int on=1;
    (void)setsockopt(msgsock,SOL_SOCKET,SO_NOSIGPIPE,&on,sizeof(on));
#endif
    sprintf(host,"%u.%u.%u.%u",
            (u_char)sadr.sa_data[2],(u_char)sadr.sa_data[3],
            (u_char)sadr.sa_data[4],(u_char)sadr.sa_data[5]);
    sprintf(buf,"connection accepted from %s",host);
    message(NULL,buf,MSG_FLUSH);
#ifdef SO_NOSIGPIPE                    /* 2017-07-05 */
    (void)setsockopt(msgsock,SOL_SOCKET,SO_NOSIGPIPE,&on,sizeof(on));
#endif
    Connection *c = malloc(sizeof(Connection));
    strcpy(c->host,host);
    c->port = port;
    c->msgsock = msgsock;
    if (tcpDebug) {                    /* allow mult. connections */
      thread_detach(run_connection,(void*)c);
    } else {                           /* single connection */
      run_connection((void*)c);
    }
  } /* while(!done) */

  (void)close(sock);

  sprintf(buf,"%s(%d) done\n",PREFUN,port);
  message(NULL,buf,MSG_FLUSH);

  return (void*)0;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

